﻿using System;
using System.Collections.ObjectModel;
using System.IO;
using System.Windows.Forms;
using ThinkGeo.Core;

namespace SplitPolygonsBasedOnGrid
{
    public partial class frmSplitPolygons : Form
    {
        private int featureCount = 0;
        private int featureIndex = 0;
        private int cellCount = 0;
        private int cellIndex = 0;
        private double miniCellArea = 0;
        private GeographyUnit sourceUnit = GeographyUnit.DecimalDegree;
        private AreaUnit miniCellUnit = AreaUnit.SquareMiles;

        public frmSplitPolygons()
        {
            InitializeComponent();
        }

        private void frmSplitPolygons_Load(object sender, EventArgs e)
        {
            txtSourceDirectory.Text = @"..\..\SampleData";

            cbSourceUnit.DataSource = Enum.GetNames(typeof(GeographyUnit));
            cbSourceUnit.SelectedIndex = 1;

            txtMinCellArea.Text = "10000";
            cbCellAreaUnit.DataSource = Enum.GetNames(typeof(AreaUnit));
            cbCellAreaUnit.SelectedIndex = 5;
        }

        private void btnLookupSource_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog folderBrowserDialog = new FolderBrowserDialog();
            folderBrowserDialog.ShowNewFolderButton = false;
            folderBrowserDialog.ShowDialog();
            txtSourceDirectory.Text = folderBrowserDialog.SelectedPath;
        }

        private void btnProcess_Click(object sender, EventArgs e)
        {
            btnProcess.Enabled = false;

            string filePath = txtSourceDirectory.Text;

            if (!Directory.Exists(filePath))
            {
                btnProcess.Enabled = true;
                MessageBox.Show("Source Directory is invalid");
                return;
            }

            if (!Double.TryParse(txtMinCellArea.Text, out miniCellArea))
            {
                btnProcess.Enabled = true;
                MessageBox.Show("Minimum Cell Area is invalid");
                return;
            }

            txtStatus.Text = "Process start";
            Application.DoEvents();

            sourceUnit = (GeographyUnit)Enum.Parse(typeof(GeographyUnit), cbSourceUnit.Text);
            miniCellUnit = (AreaUnit)Enum.Parse(typeof(AreaUnit), cbCellAreaUnit.Text);
            string[] filesToDelete = Directory.GetFiles(filePath, "*SplitTemp.*");

            foreach (string file in filesToDelete)
            {
                File.Delete(file);
            }

            string[] shapefiles = Directory.GetFiles(filePath, "*.shp", SearchOption.TopDirectoryOnly);

            foreach (string shapefile in shapefiles)
            {
                if (!shapefile.Contains("_split.shp"))
                {
                    CreateGridForShapefile(shapefile);
                }
            }

            MessageBox.Show("Finished!");
            txtStatus.Text = "Finished!";
            btnProcess.Enabled = true;
        }

        private void CreateGridForShapefile(string file)
        {
            ShapeFileFeatureSource.BuildIndexFile(file, BuildIndexMode.DoNotRebuild);
            ShapeFileFeatureSource shapeFile = new ShapeFileFeatureSource(file, FileAccess.Read);
            shapeFile.Open();
            if (shapeFile.GetShapeFileType() == ShapeFileType.Polygon)
            {
                ShapeFileFeatureSource.CreateShapeFile(ShapeFileType.Polygon, file.Replace(".shp", "_AreaSplitTemp.shp"), new Collection<DbfColumn>() { new DbfColumn("RECID", DbfColumnType.Character, 10, 0) });
                ShapeFileFeatureSource.CreateShapeFile(ShapeFileType.Polyline, file.Replace(".shp", "_LineSplitTemp.shp"), new Collection<DbfColumn>() { new DbfColumn("RECID", DbfColumnType.Character, 10, 0) });

                ShapeFileFeatureSource targetAreaShapeFile = new ShapeFileFeatureSource(file.Replace(".shp", "_AreaSplitTemp.shp"), FileAccess.ReadWrite);
                targetAreaShapeFile.Open();
                targetAreaShapeFile.BeginTransaction();

                ShapeFileFeatureSource targetLineShapeFile = new ShapeFileFeatureSource(file.Replace(".shp", "_LineSplitTemp.shp"), FileAccess.ReadWrite);
                targetLineShapeFile.Open();
                targetLineShapeFile.BeginTransaction();

                Collection<Feature> allFeatures = shapeFile.GetAllFeatures(ReturningColumnsType.AllColumns);

                featureCount = allFeatures.Count;
                featureIndex = 0;

                foreach (Feature targetFeature in allFeatures)
                {
                    featureIndex++;
                    CreateGridForFeature(shapeFile, targetAreaShapeFile, targetLineShapeFile, targetFeature);
                }

                targetLineShapeFile.CommitTransaction();
                targetAreaShapeFile.CommitTransaction();

                targetAreaShapeFile.Close();
                targetLineShapeFile.Close();

                RenameTempFiles(file.Replace(".shp", "_AreaSplitTemp.shp"));
                RenameTempFiles(file.Replace(".shp", "_LineSplitTemp.shp"));
            }
            shapeFile.Close();
        }

        private void CreateGridForFeature(ShapeFileFeatureSource shapeFile, ShapeFileFeatureSource targetAreaShapeFile, ShapeFileFeatureSource targetLineShapeFile, Feature targetFeature)
        {
            GeoCollection<Feature> splitCells = DynamicGridPolygonIndexer.GetIntersectingGridCells(targetFeature.GetShape(), sourceUnit, miniCellArea, miniCellUnit);
            cellCount = splitCells.Count;
            cellIndex = 0;

            foreach (Feature cell in splitCells)
            {
                cellIndex++;
                ProcessOneGridCell(targetAreaShapeFile, targetLineShapeFile, targetFeature, cell);
                txtStatus.Text = string.Format("Cell {0} of {1}, Feature {2} of {3} in {4}", cellIndex, cellCount, featureIndex, featureCount, Path.GetFileName(shapeFile.ShapePathFilename));
                Application.DoEvents();
            }
        }

        private void ProcessOneGridCell(ShapeFileFeatureSource targetAreaShapeFile, ShapeFileFeatureSource targetLineShapeFile, Feature targetFeature, Feature cell)
        {
            BaseShape areaIntersection = null;
            BaseShape lineIntersection = null;
            MultipolygonShape polygon = (MultipolygonShape)targetFeature.GetShape();

            areaIntersection = polygon.GetIntersection(cell.GetShape() as PolygonShape);
            MultilineShape multiLine = new MultilineShape();

            foreach (PolygonShape singlePolygon in polygon.Polygons)
            {
                multiLine.Lines.Add(new LineShape(singlePolygon.OuterRing.Vertices));

                foreach (RingShape ring in singlePolygon.InnerRings)
                {
                    multiLine.Lines.Add(new LineShape(ring.Vertices));
                }
            }

            lineIntersection = multiLine.GetIntersection(cell.GetShape() as PolygonShape);

            if (lineIntersection is MultilineShape)
            {
                if (((MultilineShape)lineIntersection).Lines.Count == 0)
                {
                    lineIntersection = null;
                }
            }

            if (areaIntersection != null)
            {
                Feature newFeature = new Feature(areaIntersection);
                newFeature.ColumnValues.Add("ID", targetFeature.Id.ToString());
                targetAreaShapeFile.AddFeature(newFeature);
            }
            if (lineIntersection != null)
            {
                Feature newFeature = new Feature(lineIntersection);
                newFeature.ColumnValues.Add("ID", targetFeature.Id.ToString());
                targetLineShapeFile.AddFeature(newFeature);
            }
        }

        private void btnCancel_Click(object sender, EventArgs e)
        {
            Environment.Exit(0);
        }

        private void RenameTempFiles(string shapePathFileName)
        {
            File.Move(shapePathFileName, shapePathFileName.Replace("SplitTemp", "Split"));
            File.Move(shapePathFileName.Replace(".shp", ".dbf"), shapePathFileName.Replace(".shp", ".dbf").Replace("SplitTemp", "Split"));
            File.Move(shapePathFileName.Replace(".shp", ".shx"), shapePathFileName.Replace(".shp", ".shx").Replace("SplitTemp", "Split"));
            File.Move(shapePathFileName.Replace(".shp", ".idx"), shapePathFileName.Replace(".shp", ".idx").Replace("SplitTemp", "Split"));
            File.Move(shapePathFileName.Replace(".shp", ".ids"), shapePathFileName.Replace(".shp", ".ids").Replace("SplitTemp", "Split"));
        }
    }
}
